#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

"""
Contains core functions for DAMON debugfs control.
"""

import os
import subprocess

import _damo_fs
import _damo_schemes_input
import _damon

debugfs = '/sys/kernel/debug'
debugfs_damon = os.path.join(debugfs, 'damon')
debugfs_version = os.path.join(debugfs_damon, 'version')
debugfs_attrs = os.path.join(debugfs_damon, 'attrs')
debugfs_record = os.path.join(debugfs_damon, 'record')
debugfs_schemes = os.path.join(debugfs_damon, 'schemes')
debugfs_target_ids = os.path.join(debugfs_damon, 'target_ids')
debugfs_init_regions = os.path.join(debugfs_damon, 'init_regions')
debugfs_monitor_on = os.path.join(debugfs_damon, 'monitor_on')

def turn_damon_on(kdamonds_names):
    return _damo_fs.write_files({debugfs_monitor_on: 'on'})

def turn_damon_off(kdamonds_names):
    return _damo_fs.write_files({debugfs_monitor_on: 'off'})

def is_kdamond_running(kdamond_name):
    content, err = _damo_fs.read_file(debugfs_monitor_on)
    if err != None:
        raise Exception('monitor_on file read failed: err')
    return content.strip() == 'on'

'Return error'
def update_schemes_stats(kdamond_names):
    # DAMON debugfs updates stats always
    return None

# update_schemes_tried_regions are not supported

# for apply_kdamonds

def wops_for_target(target, target_has_pid):
    wops = []
    if target_has_pid:
        wops.append({debugfs_target_ids: '%s' % target.pid})
        tid = target.pid
    else:
        if not feature_supported('paddr'):
            raise Exception('paddr is not supported')
        wops.append({debugfs_target_ids: 'paddr\n'})
        tid = 42
    if feature_supported('init_regions_target_idx'):
        tid = 0

    if feature_supported('init_regions'):
        string = ' '.join(['%s %d %d' % (tid, r.start, r.end) for r in
            target.regions])
        wops.append({debugfs_init_regions: string})
    return wops

damos_action_to_int = {
        _damon.damos_action_willneed: 0,
        _damon.damos_action_cold: 1,
        _damon.damos_action_pageout: 2,
        _damon.damos_action_hugepage: 3,
        _damon.damos_action_nohugepage: 4,
        _damon.damos_action_stat: 5}

damos_wmark_metric_to_int = {
        _damon.damos_wmarks_metric_none: 0,
        _damon.damos_wmarks_metric_free_mem_rate: 1}

def damos_action_to_file_input(action_str):
    return '%d' % damos_action_to_int[action_str]

def damos_wmarks_metric_to_file_input(metric_str):
    return '%d' % damos_wmark_metric_to_int[metric_str]

def file_content_to_damos_action(action_file_content):
    for action_str in damos_action_to_int:
        if damos_action_to_int[action_str] == action_file_content:
            return action_str

def file_content_to_damos_wmarks_metric(metric_file_content):
    for metric_str in damos_wmark_metric_to_int:
        if damos_wmark_metric_to_int[metric_str] == metric_file_content:
            return metric_str

def damos_to_debugfs_input(damos, intervals, scheme_version):
    pattern = damos.access_pattern.converted_for_units(
            _damon.unit_sample_intervals, _damon.unit_aggr_intervals,
            intervals)
    quotas = damos.quotas
    watermarks = damos.watermarks

    max_nr_accesses = intervals.aggr / intervals.sample
    v0_scheme = '%d\t%d\t%d\t%d\t%d\t%d\t%s' % (
            pattern.min_sz_bytes, pattern.max_sz_bytes,
            pattern.min_nr_accesses, pattern.max_nr_accesses,
            pattern.min_age, pattern.max_age,
            damos_action_to_file_input(damos.action))
    if scheme_version == 0:
        return v0_scheme

    v1_scheme = '%s\t%d\t%d' % (v0_scheme,
            quotas.sz_bytes, quotas.reset_interval_ms)
    if scheme_version == 1:
        return v1_scheme

    v2_scheme = '%s\t%d\t%d\t%d' % (v1_scheme,
            quotas.weight_sz_permil, quotas.weight_nr_accesses_permil,
            quotas.weight_age_permil)
    if scheme_version == 2:
        return v2_scheme

    v3_scheme = '%s\t%s\t%d\t%d\t%d\t%d' % (v2_scheme,
            damos_wmarks_metric_to_file_input(watermarks.metric),
            watermarks.interval_us, watermarks.high_permil,
            watermarks.mid_permil, watermarks.low_permil)
    if scheme_version == 3:
        return v3_scheme

    v4_scheme = '%s\t' % v0_scheme + '\t'.join('%d' % x for x in [quotas.time_ms,
        quotas.sz_bytes, quotas.reset_interval_ms, quotas.weight_sz_permil,
        quotas.weight_nr_accesses_permil, quotas.weight_age_permil,
        int(damos_wmarks_metric_to_file_input(watermarks.metric)),
        watermarks.interval_us, watermarks.high_permil, watermarks.mid_permil,
        watermarks.low_permil])
    if scheme_version == 4:
        return v4_scheme

    raise Exception('Unsupported scheme version: %d' % scheme_version)

def get_scheme_version():
    scheme_version = 0
    if feature_supported('schemes_speed_limit'):
        scheme_version = 1
    if feature_supported('schemes_prioritization'):
        scheme_version = 2
    if feature_supported('schemes_wmarks'):
        scheme_version = 3
    if feature_supported('schemes_quotas'):
        scheme_version = 4
    return scheme_version

def wops_for_schemes(schemes, intervals):
    scheme_file_input_lines = []
    for scheme in schemes:
        scheme_file_input_lines.append(damos_to_debugfs_input(scheme,
            intervals, get_scheme_version()))
    scheme_file_input = '\n'.join(scheme_file_input_lines)
    if scheme_file_input == '':
        scheme_file_input = '\n'
    return [{debugfs_schemes: scheme_file_input}]

def attr_str_ctx(damon_ctx):
    intervals = damon_ctx.intervals
    nr_regions = damon_ctx.nr_regions
    return '%d %d %d %d %d ' % (intervals.sample, intervals.aggr,
            intervals.ops_update, nr_regions.min_nr_regions,
            nr_regions.max_nr_regions)

def wops_for_kdamonds(kdamonds):
    if len(kdamonds) > 1:
        raise Exception('Currently only <=one kdamond is supported')
    if len(kdamonds) == 1 and len(kdamonds[0].contexts) > 1:
        raise Exception('currently only <= one damon_ctx is supported')
    if (len(kdamonds) == 1 and len(kdamonds[0].contexts) == 1 and
            len(kdamonds[0].contexts[0].targets) > 1):
        raise Exception('currently only <= one target is supported')
    ctx = kdamonds[0].contexts[0]

    write_contents = []
    write_contents.append({debugfs_attrs: attr_str_ctx(ctx)})

    if len(ctx.targets) > 0:
        write_contents += wops_for_target(ctx.targets[0],
                _damon.target_has_pid(ctx.ops))

    if feature_supported('record') and ctx.record_request != None:
        record_file_input = '%s %s' % (ctx.record_request.rfile_buf,
                ctx.record_request.rfile_path)
        write_contents.append({debugfs_record: record_file_input})

    if not debugfs_schemes:
        return write_contents

    write_contents += wops_for_schemes(ctx.schemes, ctx.intervals)

    return write_contents

def apply_kdamonds(kdamonds):
    return _damo_fs.write_files(wops_for_kdamonds(kdamonds))

# for current_kdamonds

def debugfs_output_to_damos(output, intervals_us):
    fields = [int(x) for x in output.strip().split()]
    if feature_supported('schemes_stat_succ'):
        nr_stat_fields = 5
    else:
        nr_stat_fields = 2
    stat_fields = fields[-1 * nr_stat_fields:]
    fields = [int(x) for x in output.strip().split()][:-1 * nr_stat_fields]
    # convert nr_accesses from sample intervals to percent
    max_nr_accesses = intervals_us.aggr / intervals_us.sample
    fields[2] = float(fields[2]) * 100 / max_nr_accesses
    fields[3] = float(fields[3]) * 100 / max_nr_accesses
    # convert ages in update_interval to us
    fields[4] = intervals_us.aggr * fields[4]
    fields[5] = intervals_us.aggr * fields[5]
    fields[6] = file_content_to_damos_action(fields[6])

    if len(fields) == 17:
        fields[12] = file_content_to_damos_wmarks_metric(fields[12])
    elif len(fields) == 18:
        fields[13] = file_content_to_damos_wmarks_metric(fields[13])

    line = ' '.join('%s' % x for x in fields) # remove stats
    damos, err = _damo_schemes_input.damo_single_line_scheme_to_damos(
            line, '0')
    if err != None:
        raise Exception('debugfs output to damos conversion failed (%s)' % err)
    damos.stats = _damon.DamosStats(*stat_fields)
    return damos

def files_content_to_kdamonds(files_content):
    attrs = [int(x) for x in files_content['attrs'].strip().split()]

    intervals = _damon.DamonIntervals(attrs[0], attrs[1], attrs[2])
    nr_regions = _damon.DamonNrRegionsRange(attrs[3], attrs[4])

    target_ids = [int(x) for x in files_content['target_ids'].strip().split()]
    fields = [int(x) for x in files_content['init_regions'].strip().split()]
    regions_dict = {}
    for i in range(0, len(fields), 3):
        id_or_index = fields[i]
        if not id_or_index in regions_dict:
            regions_dict[id_or_index] = []
        regions_dict[id_or_index].append(_damon.DamonRegion(
                fields[i + 1], fields[i + 2]))
    ops = 'vaddr'
    is_paddr = False
    if 42 in target_ids:
        ops = 'paddr'
    targets = []
    for idx, target_id in enumerate(target_ids):
        targets.append(_damon.DamonTarget(name='%d' % idx,
            pid=target_id if not is_paddr else None,
            regions=regions_dict[idx
                if feature_supported('init_regions_target_idx')
                else target_id] if len(regions_dict) > 0 else []))

    if feature_supported('record'):
        fields = files_content['record'].strip().split()
        record_request = _damon.DamonRecord(int(fields[0]), fields[1].strip())

    schemes = []
    for line in files_content['schemes'].split('\n'):
        if line.strip() == '':
            continue
        schemes.append(debugfs_output_to_damos(line, intervals))

    ctx = _damon.DamonCtx('0', intervals, nr_regions, ops, targets, schemes)
    if feature_supported('record'):
        ctx.record_request = record_request
    state = files_content['monitor_on'].strip()
    pid = files_content['kdamond_pid'].strip()
    return [_damon.Kdamond('0', state, pid, [ctx])]

def current_kdamonds():
    return files_content_to_kdamonds(
            _damo_fs.read_files_recursive(debugfs_damon))

def current_kdamond_names():
    # TODO: Support created kdamonds
    return ['0']

# features

feature_supports = None

def feature_supported(feature):
    if feature_supports == None:
        update_supported_features()

    return feature_supports[feature]

def values_for_restore(filepath, read_val):
    if read_val == '':
        return '\n'
    if os.path.basename(filepath) == 'target_ids' and read_val == '42\n':
        return 'paddr\n'
    return read_val

'''Return value to write back to the filepath for restoring, and error'''
def read_value_for_restore(filepath):
    err = True
    read_val, err = _damo_fs.read_file(filepath)
    if err != None:
        return None, err
    err = None
    return values_for_restore(filepath, read_val), err

def test_debugfs_file(path, input_str, expected):
    orig_val, err = read_value_for_restore(path)
    if err != None:
        return False
    err = _damo_fs.write_file(path, input_str)
    if err != None:
        return False
    content, err = _damo_fs.read_file(path)
    if err != None:
        return False
    if content == expected:
        passed = True
    else:
        passed = False
    err = _damo_fs.write_file(path, orig_val)
    if err != None:
        return False
    return passed

def test_debugfs_file_schemes(nr_fields):
    input_str = ' '.join(['1'] * nr_fields)
    expected = '%s 0 0\n' % input_str

    return test_debugfs_file(debugfs_schemes, input_str, expected)

def test_debugfs_file_schemes_stat_extended(nr_fields):
    input_str = ' '.join(['1'] * nr_fields)
    expected = '%s 0 0 0 0 0\n' % input_str

    return test_debugfs_file(debugfs_schemes, input_str, expected)

def test_init_regions_version(paddr_supported):
    # Save previous values
    orig_target_ids, err = read_value_for_restore(debugfs_target_ids)
    if err != None:
        raise Exception('debugfs target_ids read failed')
    orig_init_regions = read_value_for_restore(debugfs_init_regions)
    if err != None:
        raise Exception('debugfs init_regions read failed')

    # Test
    if paddr_supported:
        err = _damo_fs.write_file(debugfs_target_ids, 'paddr\n')
        if err != None:
            raise Exception(err)
        v1_input = '42 100 200'
    else:
        err = _damo_fs.write_file(debugfs_target_ids, '%d\n' % os.getpid())
        if err != None:
            raise Exception(err)
        v1_input = '%d 100 200' % os.getpid()

    # We check if the write was success below anyway, so ignore error
    err = _damo_fs.write_file(debugfs_init_regions, v1_input)
    read_val, err = _damo_fs.read_file(debugfs_init_regions)
    if err != None:
        raise Exception(err)
    if read_val.strip() == v1_input:
        version = 1
    else:
        version = 2

    # Previous value might be invalid now (e.g., process terminated), so ignore
    # error
    err = _damo_fs.write_file(debugfs_target_ids, orig_target_ids)
    err = _damo_fs.write_file(debugfs_init_regions, orig_init_regions)

    return version

def set_damon_debugfs_paths():
    'Set global variables for DAMON debugfs path.  Return error if failed'
    global debugfs_damon
    global debugfs_version
    global debugfs_attrs
    global debugfs_record
    global debugfs_schemes
    global debugfs_target_ids
    global debugfs_init_regions
    global debugfs_monitor_on

    if not os.path.isdir(debugfs_damon):
        return 'damon debugfs dir (%s) not found' % debugfs_damon

    for f in [debugfs_version, debugfs_attrs, debugfs_record, debugfs_schemes,
            debugfs_target_ids, debugfs_init_regions, debugfs_monitor_on]:
        # f could be None if this function is called before
        if f == None:
            continue
        if not os.path.isfile(f):
            if f == debugfs_version:
                debugfs_version = None
            elif f == debugfs_record:
                debugfs_record = None
            elif f == debugfs_schemes:
                debugfs_schemes = None
            elif f == debugfs_init_regions:
                debugfs_init_regions = None
            else:
                return 'damon debugfs file (%s) not found' % f
    return None

def update_supported_features():
    global feature_supports
    if feature_supports != None:
        return None
    feature_supports = {x: False for x in _damon.features}

    err = set_damon_debugfs_paths()
    if err != None:
        return err

    if _damon.any_kdamond_running():
        return 'debugfs feature update cannot be done while DAMON running'

    if debugfs_record != None:
        feature_supports['record'] = True
    if debugfs_schemes != None:
        feature_supports['schemes'] = True

    # virtual address space has supported since the beginning
    feature_supports['vaddr'] = True
    if test_debugfs_file(debugfs_target_ids, 'paddr\n', '42\n'):
        feature_supports['paddr'] = True

    if debugfs_init_regions != None:
        feature_supports['init_regions'] = True
        init_regions_version = test_init_regions_version(
                feature_supports['paddr'])
        if init_regions_version == 2:
            feature_supports['init_regions_target_idx'] = True

    if debugfs_schemes != None:
        if test_debugfs_file_schemes(9):
            feature_supports['schemes_speed_limit'] = True
        elif test_debugfs_file_schemes(12):
            feature_supports['schemes_speed_limit'] = True
            feature_supports['schemes_prioritization'] = True
        elif test_debugfs_file_schemes(17):
            feature_supports['schemes_speed_limit'] = True
            feature_supports['schemes_prioritization'] = True
            feature_supports['schemes_wmarks'] = True
        elif test_debugfs_file_schemes(18):
            feature_supports['schemes_speed_limit'] = True
            feature_supports['schemes_prioritization'] = True
            feature_supports['schemes_wmarks'] = True
            feature_supports['schemes_quotas'] = True
        elif test_debugfs_file_schemes_stat_extended(18):
            feature_supports['schemes_speed_limit'] = True
            feature_supports['schemes_prioritization'] = True
            feature_supports['schemes_wmarks'] = True
            feature_supports['schemes_quotas'] = True
            feature_supports['schemes_stat_succ'] = True
            feature_supports['schemes_stat_qt_exceed'] = True
    return None



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

import argparse

import _damo_schemes_input
import _damon
import _damon_dbgfs

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('input', metavar='<file or schemes in text>',
            help='input file describing the schemes or the schemes')
    parser.add_argument('-s', '--sample', metavar='<interval>', type=int,
            default=5000, help='sampling interval (us)')
    parser.add_argument('-a', '--aggr', metavar='<interval>', type=int,
            default=100000, help='aggregation interval (us)')
    parser.add_argument('--scheme_version', metavar='<version>', type=int,
            choices=range(0, 5), default=4, help='destination scheme version')
    args = parser.parse_args()

    intervals = _damon.DamonIntervals(args.sample, args.aggr, args.aggr)

    lines = []
    damos_list, err = _damo_schemes_input.damo_schemes_to_damos(args.input)
    if err:
        print('cannot create damos from argument (%s)' % err)
        exit(1)
    for damos in damos_list:
        lines.append(
                _damon_dbgfs.damos_to_debugfs_input(damos, intervals,
                    args.scheme_version))
    print ('\n'.join(lines))

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

"""
Record monitored data access patterns.
"""

import argparse
import os
import signal
import subprocess
import time

import _damon
import _damon_args
import _damon_result
import _damo_paddr_layout

class DataForCleanup:
    kdamonds_names = None
    target_is_ongoing = False
    orig_kdamonds = None
    rfile_path = None
    rfile_format = None
    rfile_permission = None
    remove_perf_data = False
    perf_pipe = None

data_for_cleanup = DataForCleanup()

def change_rfile_format(rfile_path, src_format, dst_format, dst_permission):
    rfile_path_mid = rfile_path + '.mid'
    os.rename(rfile_path, rfile_path_mid)
    result, err = _damon_result.parse_damon_result(rfile_path_mid, src_format)
    if err:
        raise Exception('parsing DAMON output failed (%s)' % err)

    _damon_result.write_damon_result(result, rfile_path, dst_format,
            dst_permission)
    os.remove(rfile_path_mid)

def cleanup_exit(exit_code):
    if data_for_cleanup.perf_pipe:
        try:
            data_for_cleanup.perf_pipe.send_signal(signal.SIGINT)
            data_for_cleanup.perf_pipe.wait()
        except:
            # perf might already finished
            pass

        rfile_current_format = 'perf_script'
        perf_data = data_for_cleanup.rfile_path + '.perf.data'
        subprocess.call('perf script -i \'%s\' > \'%s\'' %
                (perf_data, data_for_cleanup.rfile_path),
                shell=True, executable='/bin/bash')
        if data_for_cleanup.remove_perf_data:
            os.remove(perf_data)
    else:
        rfile_current_format = 'record'

    if not data_for_cleanup.target_is_ongoing:
        if _damon.any_kdamond_running():
            if data_for_cleanup.kdamonds_names == None:
                # turn on failed
                pass
            else:
                err = _damon.turn_damon_off(data_for_cleanup.kdamonds_names)
                if err:
                    print('failed to turn damon off (%s)' % err)
        err = _damon.apply_kdamonds(data_for_cleanup.orig_kdamonds)
        if err:
            print('failed restoring previous kdamonds setup (%s)' % err)

    if (data_for_cleanup.rfile_format != None and
            rfile_current_format != data_for_cleanup.rfile_format):
        change_rfile_format(data_for_cleanup.rfile_path, rfile_current_format,
                data_for_cleanup.rfile_format,
                data_for_cleanup.rfile_permission)

    # DAMON might not started at all
    if os.path.isfile(data_for_cleanup.rfile_path):
        os.chmod(data_for_cleanup.rfile_path,
                data_for_cleanup.rfile_permission)

    exit(exit_code)

def sighandler(signum, frame):
    print('\nsignal %s received' % signum)
    cleanup_exit(signum)

def set_data_for_cleanup(data_for_cleanup, args, output_permission):
    data_for_cleanup.target_is_ongoing = _damon_args.is_ongoing_target(args)
    data_for_cleanup.rfile_format = args.output_type
    data_for_cleanup.rfile_path = args.out
    data_for_cleanup.remove_perf_data = not args.leave_perf_data
    data_for_cleanup.rfile_permission = output_permission
    data_for_cleanup.orig_kdamonds = _damon.current_kdamonds()

def chk_handle_record_feature_support(args):
    damon_record_supported = _damon.feature_supported('record')
    if not damon_record_supported:
        try:
            subprocess.check_output(['which', 'perf'])
        except:
            print('perf is not installed')
            exit(1)
        if args.rbuf:
            print('# \'--rbuf\' will be ignored')

    if not args.rbuf:
        args.rbuf = 1024 * 1024

    return damon_record_supported

def chk_handle_output_permission(output_permission_option):
    output_permission = int(output_permission_option, 8)
    if output_permission < 0o0 or output_permission > 0o777:
        print('wrong --output_permission (%s)' % output_permission_option)
        exit(1)
    return output_permission

def backup_duplicate_output_file(output_file):
    if os.path.isfile(output_file):
        os.rename(output_file, output_file + '.old')

def set_argparser(parser):
    _damon_args.set_implicit_target_record_argparser(parser)
    parser.add_argument('--output_type', choices=['record', 'perf_script'],
            default=None, help='output file\'s type')
    parser.add_argument('--leave_perf_data', action='store_true',
            default=False, help='don\'t remove the perf.data file')
    parser.add_argument('--output_permission', type=str, default='600',
            help='permission of the output file')

def main(args=None):
    global data_for_cleanup

    if not args:
        parser = argparse.ArgumentParser()
        set_argparser(parser)
        args = parser.parse_args()

    _damon.ensure_root_and_initialized(args)

    # Check/handle the arguments and options
    damon_record_supported = chk_handle_record_feature_support(args)
    output_permission = chk_handle_output_permission(args.output_permission)
    backup_duplicate_output_file(args.out)

    # Setup for cleanup
    set_data_for_cleanup(data_for_cleanup, args, output_permission)
    signal.signal(signal.SIGINT, sighandler)
    signal.signal(signal.SIGTERM, sighandler)

    # Now the real works
    if not _damon_args.is_ongoing_target(args):
        # Turn DAMON on
        err, kdamonds = _damon_args.turn_implicit_args_damon_on(args)
        if err:
            print('could not turn DAMON on (%s)' % err)
            cleanup_exit(-2)
        data_for_cleanup.kdamonds_names = [k.name for k in kdamonds]


    if not damon_record_supported:
        # Record the monitoring results using perf
        data_for_cleanup.perf_pipe = subprocess.Popen([
            'perf', 'record', '-a', '-e', 'damon:damon_aggregated',
            '-o', data_for_cleanup.rfile_path + '.perf.data'])
    print('Press Ctrl+C to stop')

    if not _damon_args.is_ongoing_target(args) and args.self_started_target == True:
        os.waitpid(kdamonds[0].contexts[0].targets[0].pid, 0)
    _damon.wait_current_kdamonds_turned_off()

    cleanup_exit(0)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import argparse
import json

import _damon
import _damon_args

def main():
    parser = argparse.ArgumentParser()
    _damon_args.set_explicit_target_argparser(parser)
    args = parser.parse_args()

    _damon.ensure_root_and_initialized(args)

    kdamonds, err = _damon_args.kdamonds_from_damon_args(args)
    if err:
        print('creating kdamonds from arguments failed (%s)' % err)
        exit(1)
    kvpairs = [k.to_kvpairs() for k in kdamonds]
    print(json.dumps(kvpairs, indent=4, sort_keys=True))

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

import argparse

import damo_stat

import _damo_fmt_str
import _damo_subcmds
import _damon

def pr_schemes_tried_regions(kdamond_name, monitoring_scheme, raw_nr):
    for kdamond in _damon.current_kdamonds():
        if kdamond.name != kdamond_name:
            continue
        for ctx in kdamond.contexts:
            for scheme in ctx.schemes:
                if scheme == monitoring_scheme:
                    print('\n'.join(r.to_str(raw_nr, ctx.intervals) for r in
                        scheme.tried_regions))
                    return

def monitoring_kdamond_scheme():
    monitoring_kdamond = None
    monitoring_scheme = None
    kdamonds = _damon.current_kdamonds()
    for kdamond in kdamonds:
        for ctx in kdamond.contexts:
            for scheme in ctx.schemes:
                if _damon.is_monitoring_scheme(scheme, ctx.intervals):
                    return kdamond.name, scheme
    return None, None

def update_pr_schemes_tried_regions(raw_nr):
    if _damon.every_kdamond_turned_off():
        print('no kdamond running')
        exit(1)

    monitoring_kdamond, monitoring_scheme = monitoring_kdamond_scheme()
    if monitoring_kdamond == None:
        print('no kdamond is having monitoring scheme')
        exit(1)

    err = _damon.update_schemes_tried_regions([monitoring_kdamond])
    if err != None:
        print('update schemes tried regions fail: %s', err)
        exit(1)

    pr_schemes_tried_regions(monitoring_kdamond, monitoring_scheme, raw_nr)

def set_argparser(parser):
    damo_stat.set_common_argparser(parser)

def __main(args):
    if not _damon.feature_supported('schemes_tried_regions'):
        print('schemes_tried_regions feature not supported')
        exit(1)
    update_pr_schemes_tried_regions(args.raw)

def main(args=None):
    if not args:
        parser = argparse.ArgumentParser()
        set_argparser(parser)
        args = parser.parse_args()

    _damon.ensure_root_and_initialized(args)

    damo_stat.run_count_delay(__main, args)

    for i in range(args.count):
        if i != args.count - 1:
            time.sleep(args.delay)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

"""
Start DAMON with given parameters.
"""

import argparse

import _damon
import _damon_args
import _damo_paddr_layout

def set_argparser(parser):
    _damon_args.set_explicit_target_argparser(parser)

def main(args=None):
    if not args:
        parser = argparse.ArgumentParser()
        set_argparser(parser)
        args = parser.parse_args()

    _damon.ensure_root_and_initialized(args)

    err, kdamonds = _damon_args.turn_explicit_args_damon_on(args)
    if err:
        print('could not turn on damon (%s)' % err)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

import argparse
import json

import _damon
import _damon_args

def set_argparser(parser):
    parser.add_argument('operation', choices=['read', 'write'],
            help='operation to do')
    parser.add_argument('--content', help='content to write')
    _damon_args.set_common_argparser(parser)

def main(args=None):
    if not args:
        parser = argparse.ArgumentParser()
        set_argparser(parser)
        args = parser.parse_args()

    _damon.ensure_root_and_initialized(args)

    if args.operation == 'read':
        content = _damon.read_damon_fs()
        print(json.dumps(content, indent=4, sort_keys=True))
    if args.operation == 'write':
        if args.content == None:
            print('\'--content\' should provided for write')
            exit(1)
        err = _damon.write_damon_fs(json.loads(args.content))
        if err:
            print(err)
            exit(1)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

"""
Contains core functions for DAMON sysfs control.
"""

import os
import time
import traceback

import _damo_fs
import _damon
import _damon_result

feature_supports = None

# Use only one kdamond, one context, and one target for now
root_dir = '/sys/kernel/mm/damon'
admin_dir = os.path.join(root_dir, 'admin')
kdamonds_dir = os.path.join(admin_dir, 'kdamonds')
nr_kdamonds_file = os.path.join(kdamonds_dir, 'nr_kdamonds')

def kdamond_dir_of(kdamond_name):
    return os.path.join(admin_dir, 'kdamonds', '%s' % kdamond_name)

def state_file_of(kdamond_name):
    return os.path.join(kdamond_dir_of(kdamond_name), 'state')

def nr_contexts_file_of(kdamond_name):
    return os.path.join(
            kdamond_dir_of(kdamond_name), 'contexts', 'nr_contexts')

def ctx_dir_of(kdamond_name, context_name):
    return os.path.join(
            kdamond_dir_of(kdamond_name), 'contexts', '%s' % context_name)

def schemes_dir_of(kdamond_name, context_name):
    return os.path.join(ctx_dir_of(kdamond_name, context_name), 'schemes')

def nr_schemes_file_of(kdamond_name, context_name):
    return os.path.join(
            schemes_dir_of(kdamond_name, context_name), 'nr_schemes')

def scheme_dir_of(kdamond_name, context_name, scheme_name):
    return os.path.join(
            schemes_dir_of(kdamond_name, context_name), '%s' % scheme_name)

def filters_dir_of(kdamond_name, context_name, scheme_name):
    return os.path.join(
            scheme_dir_of(kdamond_name, context_name, scheme_name), 'filters')

def nr_filters_file_of(kdamond_name, context_name, scheme_name):
    return os.path.join(
            filters_dir_of(kdamond_name, context_name, scheme_name),
            'nr_filters')

def targets_dir_of(kdamond_name, context_name):
    return os.path.join(ctx_dir_of(kdamond_name, context_name), 'targets')

def nr_targets_file_of(kdamond_name, context_name):
    return os.path.join(
            targets_dir_of(kdamond_name, context_name), 'nr_targets')

def target_dir_of(kdamond_name, context_name, target_name):
    return os.path.join(ctx_dir_of(kdamond_name, context_name), 'targets',
            '%s' % target_name)

def regions_dir_of(kdamond_name, context_name, target_name):
    return os.path.join(
            target_dir_of(kdamond_name, context_name, target_name), 'regions')

def nr_regions_file_of(kdamond_name, context_name, target_name):
    return os.path.join(
            regions_dir_of(kdamond_name, context_name, target_name),
            'nr_regions')

def supported():
    return os.path.isdir(kdamonds_dir)

def turn_damon_on(kdamonds_names):
    # In case of vaddr, too early monitoring shows unstable mapping changes.
    # Give the process a time to have stable memory mapping.
    time.sleep(0.5)
    for kdamond_name in kdamonds_names:
        err = _damo_fs.write_file(state_file_of(kdamond_name), 'on')
        if err != None:
            return err
    return None

def turn_damon_off(kdamonds_names):
    for kdamond_name in kdamonds_names:
        err = _damo_fs.write_file(state_file_of(kdamond_name), 'off')
        if err != None:
            return err
    return None

def is_kdamond_running(kdamond_name):
    content, err = _damo_fs.read_file(state_file_of(kdamond_name))
    if err != None:
        print(err)
        return False
    return content.strip() == 'on'

'Return error'
def update_schemes_stats(kdamond_names):
    for kdamond_name in kdamond_names:
        err = _damo_fs.write_files({state_file_of(kdamond_name):
            'update_schemes_stats'})
        if err != None:
            return err
    return None

'Return error'
def update_schemes_tried_regions(kdamond_names):
    for kdamond_name in kdamond_names:
        err = _damo_fs.write_files({state_file_of(kdamond_name):
            'update_schemes_tried_regions'})
        if err != None:
            return err
    return None

# for apply_kdamonds

def wops_for_scheme_filter(damos_filter):
    return {
        'type': '%s' % damos_filter.filter_type,
        'memcg_path': ('%s' % damos_filter.memcg_path
            if damos_filter.memcg_path != None else ''),
        'matching': 'Y' if damos_filter.matching else 'N',
        }

def wops_for_scheme_filters(filters):
    wops = {}
    for damos_filter in filters:
        wops[damos_filter.name] = wops_for_scheme_filter(damos_filter)
    return wops

def wops_for_scheme_watermarks(wmarks):
    if wmarks == None:
        return {}
    return {
        'metric': wmarks.metric,
        'interval_us': '%d' % wmarks.interval_us,
        'high': '%d' % wmarks.high_permil,
        'mid': '%d' % wmarks.mid_permil,
        'low': '%d' % wmarks.low_permil,
    }

def wops_for_scheme_quotas(quotas):
    if quotas == None:
        return {}
    return {
        'ms': '%d' % quotas.time_ms,
        'bytes': '%d' % quotas.sz_bytes,
        'reset_interval_ms': '%d' % quotas.reset_interval_ms,
        'weights': {
            'sz_permil': '%d' % quotas.weight_sz_permil,
            'nr_accesses_permil': '%d' % quotas.weight_nr_accesses_permil,
            'age_permil': '%d' % quotas.weight_age_permil,
        },
    }

def wops_for_scheme_access_pattern(pattern, ctx):
    if pattern == None:
        return {}
    pattern = pattern.converted_for_units(
            _damon.unit_sample_intervals, _damon.unit_aggr_intervals,
            ctx.intervals)

    return {
        'sz': {
            'min': '%d' % pattern.min_sz_bytes,
            'max': '%d' % pattern.max_sz_bytes,
        },
        'nr_accesses': {
            'min': '%d' % pattern.min_nr_accesses,
            'max': '%d' % pattern.max_nr_accesses,
        },
        'age': {
            'min': '%d' % pattern.min_age,
            'max': '%d' % pattern.max_age,
        },
    }

def wops_for_schemes(ctx):
    schemes = ctx.schemes

    schemes_wops = {}
    for scheme in schemes:
        schemes_wops[scheme.name] = {
            'access_pattern': wops_for_scheme_access_pattern(
                scheme.access_pattern, ctx),
            'action': scheme.action,
            'quotas': wops_for_scheme_quotas(scheme.quotas),
            'watermarks': wops_for_scheme_watermarks(scheme.watermarks),
        }
        if feature_supported('schemes_filters'):
            schemes_wops[scheme.name]['filters'] = wops_for_scheme_filters(
                    scheme.filters)
    return schemes_wops

def wops_for_regions(regions):
    return {'%d' % region_idx: {
        'start': '%d' % region.start,
        'end': '%d' % region.end}
        for region_idx, region in enumerate(regions)}

def wops_for_targets(ctx):
    return {
            target.name: {
                'pid_target': '%s' %
                target.pid if _damon.target_has_pid(ctx.ops) else '',
                'regions': wops_for_regions(target.regions)
                } for target in ctx.targets}

def wops_for_monitoring_attrs(ctx):
    return {
        'intervals': {
            'sample_us': '%d' % ctx.intervals.sample,
            'aggr_us': '%d' % ctx.intervals.aggr,
            'update_us': '%d' % ctx.intervals.ops_update,
        },
        'nr_regions': {
            'min': '%d' % ctx.nr_regions.min_nr_regions,
            'max': '%d' % ctx.nr_regions.max_nr_regions,
        },
    }

def wops_for_ctx(ctx):
    return [
            {'operations': ctx.ops},
            {'monitoring_attrs': wops_for_monitoring_attrs(ctx)},
            {'targets': wops_for_targets(ctx)},
            {'schemes': wops_for_schemes(ctx)},
    ]

def wops_for_ctxs(ctxs):
    return {ctx.name: wops_for_ctx(ctx) for ctx in ctxs}

def wops_for_kdamond(kdamond):
    return {'contexts': wops_for_ctxs(kdamond.contexts)}

def wops_for_kdamonds(kdamonds):
    return {kdamond.name: wops_for_kdamond(kdamond) for kdamond in kdamonds}

def ensure_dirs_populated_for(kdamonds):
    nr_kdamonds, err = _damo_fs.read_file(nr_kdamonds_file)
    if err != None:
        return err
    if int(nr_kdamonds) != len(kdamonds):
        _damo_fs.write_file_ensure(nr_kdamonds_file, '%d' % len(kdamonds))
    for kdamond in kdamonds:
        nr_contexts, err = _damo_fs.read_file(
                nr_contexts_file_of(kdamond.name))
        if err != None:
            return err
        if int(nr_contexts) != len(kdamond.contexts):
            _damo_fs.write_file_ensure(nr_contexts_file_of(kdamond.name),
                    '%d' % len(kdamond.contexts))
        for ctx in kdamond.contexts:
            nr_targets, err = _damo_fs.read_file(
                    nr_targets_file_of(kdamond.name, ctx.name))
            if err != None:
                return err
            if int(nr_targets) != len(ctx.targets):
                _damo_fs.write_file_ensure(
                        nr_targets_file_of(kdamond.name, ctx.name),
                        '%d' % len(ctx.targets))
            for target in ctx.targets:
                nr_regions, err = _damo_fs.read_file(
                        nr_regions_file_of(
                            kdamond.name, ctx.name, target.name))
                if err != None:
                    return err
                if int(nr_regions) != len(target.regions):
                    _damo_fs.write_file_ensure(
                            nr_regions_file_of(
                                kdamond.name, ctx.name, target.name),
                            '%d' % len(target.regions))
        nr_schemes, err = _damo_fs.read_file(
                nr_schemes_file_of(kdamond.name, ctx.name))
        if err != None:
            return err
        if int(nr_schemes) != len(ctx.schemes):
            _damo_fs.write_file_ensure(
                    nr_schemes_file_of(kdamond.name, ctx.name),
                    '%d' % len(ctx.schemes))
        for scheme in ctx.schemes:
            nr_filters, err = _damo_fs.read_file(
                    nr_filters_file_of(kdamond.name, ctx.name, scheme.name))
            if err != None:
                return err
            if int(nr_filters) != len(scheme.filters):
                _damo_fs.write_file_ensure(
                        nr_filters_file_of(kdamond.name, ctx.name,
                            scheme.name),
                        '%d' % len(scheme.filters))

def apply_kdamonds(kdamonds):
    if len(kdamonds) > 1:
        return 'currently only <=one kdamond is supported'
    if len(kdamonds) == 1 and len(kdamonds[0].contexts) > 1:
        return 'currently only <=one damon_ctx is supported'
    if (len(kdamonds) == 1 and len(kdamonds[0].contexts) == 1 and
            len(kdamonds[0].contexts[0].targets) > 1):
        return 'currently only <=one target is supported'
    ensure_dirs_populated_for(kdamonds)

    return _damo_fs.write_files({kdamonds_dir: wops_for_kdamonds(kdamonds)})

# for current_kdamonds()

def files_content_to_access_pattern(files_content):
    return _damon.DamosAccessPattern(
            int(files_content['sz']['min']),
            int(files_content['sz']['max']),
            int(files_content['nr_accesses']['min']),
            int(files_content['nr_accesses']['max']),
            _damon.unit_sample_intervals, # nr_accesses_unit
            int(files_content['age']['min']),
            int(files_content['age']['max']),
            _damon.unit_aggr_intervals) # age_unit

def files_content_to_quotas(files_content):
    return _damon.DamosQuotas(
            int(files_content['ms']),
            int(files_content['bytes']),
            int(files_content['reset_interval_ms']),
            int(files_content['weights']['sz_permil']),
            int(files_content['weights']['nr_accesses_permil']),
            int(files_content['weights']['age_permil']))

def files_content_to_watermarks(files_content):
    return _damon.DamosWatermarks(
            files_content['metric'].strip(),
            int(files_content['interval_us']),
            int(files_content['high']),
            int(files_content['mid']),
            int(files_content['low']))

def files_content_to_damos_filters(files_content):
    filters = []
    for filter_name in files_content:
        if filter_name == 'nr_filters':
            continue
        filter_kv = files_content[filter_name]
        filters.append(_damon.DamosFilter(filter_name,
            filter_kv['type'].strip(), filter_kv['memcg_path'].strip(),
            filter_kv['matching'].strip()))
    return filters

def files_content_to_damos_stats(files_content):
    return _damon.DamosStats(
            int(files_content['nr_tried']),
            int(files_content['sz_tried']),
            int(files_content['nr_applied']),
            int(files_content['sz_applied']),
            int(files_content['qt_exceeds']))

def files_content_to_damos_tried_regions(files_content):
    regions = []
    nr_region_dirs = len(files_content)
    if 'sz_regions_sum' in files_content:
        nr_region_dirs -= 1
    for i in range(nr_region_dirs):
        regions.append(_damon.DamosTriedRegion(
            int(files_content['%d' % i]['start']),
            int(files_content['%d' % i]['end']),
            int(files_content['%d' % i]['nr_accesses']),
            int(files_content['%d' % i]['age']),
            ))
    return regions

def files_content_to_scheme(scheme_name, files_content):
    return _damon.Damos(scheme_name,
            files_content_to_access_pattern(files_content['access_pattern']),
            files_content['action'].strip(),
            files_content_to_quotas(files_content['quotas']),
            files_content_to_watermarks(files_content['watermarks']),
            files_content_to_damos_filters(files_content['filters'])
                if 'filters' in files_content else [],
            files_content_to_damos_stats(files_content['stats']),
            files_content_to_damos_tried_regions(
                files_content['tried_regions'])
                if 'tried_regions' in files_content else None)

def files_content_to_regions(files_content):
    regions = []
    for region_idx in range(int(files_content['nr_regions'])):
        region_name = '%d' % region_idx
        regions.append(_damon.DamonRegion(
            int(files_content[region_name]['start']),
            int(files_content[region_name]['end'])))
    return regions

def files_content_to_target(target_name, files_content):
    try:
        pid = int(files_content['pid_target'])
    except:
        pid = None
    regions = files_content_to_regions(files_content['regions'])
    return _damon.DamonTarget(target_name, pid, regions)

def files_content_to_context(context_name, files_content):
    mon_attrs_content = files_content['monitoring_attrs']
    intervals_content = mon_attrs_content['intervals']
    intervals = _damon.DamonIntervals(
            int(intervals_content['sample_us']),
            int(intervals_content['aggr_us']),
            int(intervals_content['update_us']))
    nr_regions_content = mon_attrs_content['nr_regions']
    nr_regions = _damon.DamonNrRegionsRange(
            int(nr_regions_content['min']),
            int(nr_regions_content['max']))
    ops = files_content['operations'].strip()

    targets_content = files_content['targets']
    targets = []
    for target_name in targets_content:
        if target_name == 'nr_targets':
            continue
        targets.append(files_content_to_target(target_name,
            targets_content[target_name]))

    schemes_content = files_content['schemes']
    schemes = []
    for scheme_name in schemes_content:
        if scheme_name == 'nr_schemes':
            continue
        schemes.append(files_content_to_scheme(scheme_name,
            schemes_content[scheme_name]))

    return _damon.DamonCtx(context_name, intervals, nr_regions, ops, targets,
            schemes)

def files_content_to_kdamond(kdamond_name, files_content):
    contexts_content = files_content['contexts']
    contexts = []
    for ctx_name in contexts_content:
        if ctx_name == 'nr_contexts':
            continue
        contexts.append(files_content_to_context(ctx_name,
            contexts_content[ctx_name]))
    state = files_content['state'].strip()
    pid = files_content['pid'].strip()
    return _damon.Kdamond(kdamond_name, state, pid, contexts)

def files_content_to_kdamonds(files_contents):
    kdamonds = []
    for kdamond_name in files_contents:
        if kdamond_name == 'nr_kdamonds':
            continue
        kdamonds.append(files_content_to_kdamond(
            kdamond_name, files_contents[kdamond_name]))
    return kdamonds

def current_kdamonds():
    return files_content_to_kdamonds(
            _damo_fs.read_files_recursive(kdamonds_dir))

def current_kdamond_names():
    # TODO: Do not read recursive but just one depth
    return [x for x in _damo_fs.read_files_recursive(kdamonds_dir).keys()
            if x != 'nr_kdamonds']

def commit_inputs(kdamonds):
    for kdamond in kdamonds:
        err = _damo_fs.write_file(state_file_of(kdamond.name), 'commit')
        if err != None:
            return err
    return None

# features

def feature_supported(feature):
    if feature_supports == None:
        update_supported_features()
    return feature_supports[feature]

features_sysfs_support_from_begining = [
        'schemes',
        'init_regions',
        'vaddr',
        'fvaddr',
        'paddr',
        'init_regions_target_idx',
        'schemes_speed_limit',
        'schemes_quotas',
        'schemes_prioritization',
        'schemes_wmarks',
        'schemes_stat_succ',
        'schemes_stat_qt_exceed',
        ]

def _avail_ops():
    '''Assumes called by update_supported_features() assuming one scheme.
    Returns available ops input and error'''
    avail_ops = []
    avail_operations_filepath = os.path.join(ctx_dir_of(0, 0),
            'avail_operations')
    if not os.path.isfile(avail_operations_filepath):
        operations_filepath = os.path.join(ctx_dir_of(0, 0), 'operations')
        for ops in ['vaddr', 'paddr', 'fvaddr']:
            err = _damo_fs.write_file(operations_filepath, ops)
            if err != None:
                avail_ops.append(ops)
        return avail_ops, None

    content, err = _damo_fs.read_file(avail_operations_filepath)
    if err != None:
        return None, err
    return content.strip().split(), None

def update_supported_features():
    global feature_supports

    if feature_supports != None:
        return None
    feature_supports = {x: False for x in _damon.features}

    if not supported():
        return 'damon sysfs dir (%s) not found' % kdamonds_dir
    for feature in features_sysfs_support_from_begining:
        feature_supports[feature] = True

    orig_kdamonds = None
    if not os.path.isdir(scheme_dir_of(0, 0, 0)):
        orig_kdamonds = current_kdamonds()
        kdamonds_for_feature_check = [_damon.Kdamond(name='0', state=None,
            pid=None, contexts=[_damon.DamonCtx(name='0', intervals=None,
                nr_regions=None, ops=None, targets=[],
                schemes=[_damon.Damos(name='0', access_pattern=None,
                    action='stat', quotas=None, watermarks=None, filters=[],
                    stats=None)])])]
        ensure_dirs_populated_for(kdamonds_for_feature_check)

    if os.path.isdir(os.path.join(scheme_dir_of(0, 0, 0), 'tried_regions')):
        feature_supports['schemes_tried_regions'] = True

    if os.path.isdir(os.path.join(scheme_dir_of(0, 0, 0), 'filters')):
        feature_supports['schemes_filters'] = True

    avail_ops, err = _avail_ops()
    if err == None:
        for ops in ['vaddr', 'paddr', 'fvaddr']:
            feature_supports[ops] = ops in avail_ops
    if orig_kdamonds != None:
        err = apply_kdamonds(orig_kdamonds)
    return err



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

"""
Update DAMON input parameters.
"""

import argparse

import _damon
import _damon_args

def set_argparser(parser):
    _damon_args.set_explicit_target_argparser(parser)

def main(args=None):
    if not args:
        parser = argparse.ArgumentParser()
        set_argparser(parser)
        args = parser.parse_args()

    _damon.ensure_root_and_initialized(args)

    if _damon.damon_interface() == 'debugfs':
        print('tune does not support debugfs interface')
        exit(1)

    if _damon.every_kdamond_turned_off():
        print('DAMON is not turned on')
        exit(1)

    kdamonds, err = _damon_args.apply_explicit_args_damon(args)
    if err:
        print('could not apply inputs (%s)' % err)
    err = _damon.commit_inputs(kdamonds)
    if err:
        print('could not commit inputs (%s)' % err)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

"""
Command line arguments handling
"""

import json
import subprocess

import _damo_schemes_input
import _damo_paddr_layout
import _damon

# Kdamonds construction from command line arguments

def damos_from_args(args):
    schemes = []
    if not 'schemes' in args or args.schemes == None:
        return schemes, None

    schemes, err = _damo_schemes_input.damo_schemes_to_damos(args.schemes)
    if err:
        return None, 'failed damo schemes arguents parsing (%s)' % err
    return schemes, None

def init_regions_from_damon_args(args):
    init_regions = []
    if args.regions:
        for region in args.regions.split():
            addrs = region.split('-')
            try:
                if len(addrs) != 2:
                    raise Exception ('two addresses not given')
                region = _damon.DamonRegion(addrs[0], addrs[1])
                if region.start >= region.end:
                    raise Exception('start >= end')
                if init_regions and init_regions[-1].end > region.start:
                    raise Exception('regions overlap')
            except Exception as e:
                return None, 'Wrong \'--regions\' argument (%s)' % e
            init_regions.append(region)

    if args.ops == 'paddr' and not init_regions:
        if args.numa_node != None:
            init_regions = _damo_paddr_layout.paddr_region_of(args.numa_node)
        else:
            init_regions = [_damo_paddr_layout.default_paddr_region()]
        try:
            init_regions = [_damon.DamonRegion(r[0], r[1])
                    for r in init_regions]
        except Exception as e:
            return None, 'Wrong \'--regions\' argument (%s)' % e

    return init_regions, None

def damon_ctx_from_damon_args(args):
    try:
        intervals = _damon.DamonIntervals(args.sample, args.aggr, args.updr)
    except Exception as e:
        return None, 'invalid intervals arguments (%s)' % e
    try:
        nr_regions = _damon.DamonNrRegionsRange(args.minr, args.maxr)
    except Exception as e:
        return None, 'invalid nr_regions arguments (%s)' % e
    ops = args.ops
    if not _damon.feature_supported(ops):
        return None, '%s unsupported' % ops

    init_regions, err = init_regions_from_damon_args(args)
    if err:
        return None, err

    try:
        target = _damon.DamonTarget('0', args.target_pid
                if _damon.target_has_pid(ops) else None, init_regions)
    except Exception as e:
        return 'Wrong \'--target_pid\' argument (%s)' % e

    record_request = None
    if _damon.feature_supported('record') and 'rbuf' in args:
        try:
            record_request = _damon.DamonRecord(args.rbuf, args.out)
        except Exception as e:
            return 'Wrong record arguments (%s)' % e

    schemes, err = damos_from_args(args)
    if err:
        return err

    try:
        ctx = _damon.DamonCtx('0', intervals, nr_regions, ops, [target],
                schemes, record_request)
        return ctx, None
    except Exception as e:
        return None, 'Creating context from arguments failed (%s)' % e

def uncomment_kvpairs_str(string):
    lines = []
    for line in string.split('\n'):
        if line.strip().startswith('#'):
            continue
        lines.append(line)
    return '\n'.join(lines).strip()

def kdamonds_from_damon_args(args):
    if args.kdamonds:
        if os.path.isfile(args.kdamonds):
            with open(args.kdamonds, 'r') as f:
                kdamonds_str = f.read()
        else:
            kdamonds_str = args.kdamonds
        kdamonds_kvpairs = json.loads(uncomment_kvpairs_str(kdamonds_str))
        return [kvpairs_to_Kdamond(kvpair)
                for kvpair in kdamonds_kvpairs], None
    ctx, err = damon_ctx_from_damon_args(args)
    if err:
        return None, err
    return [_damon.Kdamond(name='0', state=None, pid=None,
        contexts=[ctx])], None

def set_implicit_target_args_explicit(args):
    args.kdamonds = None
    args.self_started_target = False
    if args.target == 'paddr':
        args.ops = 'paddr'
        args.target_pid = None
        return None
    if not subprocess.call('which %s &> /dev/null' % args.target.split()[0],
            shell=True, executable='/bin/bash'):
        p = subprocess.Popen(args.target, shell=True, executable='/bin/bash')
        args.ops = 'vaddr'
        args.target_pid = p.pid
        args.self_started_target = True
        if args.regions and _damon.feature_supported('fvaddr'):
            args.ops = 'fvaddr'
        return None
    try:
        pid = int(args.target)
    except:
        return 'target \'%s\' is not supported' % args.target
    args.ops = 'vaddr'
    args.target_pid = pid
    if args.regions and _damon.feature_supported('fvaddr'):
        args.ops = 'fvaddr'

    return None

# Command line processing helpers

def is_ongoing_target(args):
    return args.target == 'ongoing'

def apply_explicit_args_damon(args):
    kdamonds, err = kdamonds_from_damon_args(args)
    if err:
        return None, 'cannot create kdamonds from args'
    err = _damon.apply_kdamonds(kdamonds)
    if err:
        return None, err
    return kdamonds, None

def turn_explicit_args_damon_on(args):
    kdamonds, err = apply_explicit_args_damon(args)
    if err:
        return err, None
    return _damon.turn_damon_on(
            [k.name for k in kdamonds]), kdamonds

def turn_implicit_args_damon_on(args):
    err = set_implicit_target_args_explicit(args)
    if err:
        return err, None
    return turn_explicit_args_damon_on(args)

# Commandline options setup helpers

def set_common_argparser(parser):
    parser.add_argument('--damon_interface',
            choices=['debugfs', 'sysfs', 'auto'],
            default='auto', help='underlying DAMON interface to use')
    parser.add_argument('--debug_damon', action='store_true',
            help='Print debugging log')

def set_common_monitoring_argparser(parser):
    parser.add_argument('-s', '--sample', metavar='<interval>',
            default=5000, help='sampling interval (us)')
    parser.add_argument('-a', '--aggr', metavar='<interval>',
            default=100000, help='aggregate interval (us)')
    parser.add_argument('-u', '--updr', metavar='<interval>',
            default=1000000, help='regions update interval (us)')
    parser.add_argument('-n', '--minr', metavar='<# regions>',
            default=10, help='minimal number of regions')
    parser.add_argument('-m', '--maxr', metavar='<# regions>',
            default=1000, help='maximum number of regions')
    parser.add_argument('-r', '--regions', metavar='"<start>-<end> ..."',
            type=str, default='', help='monitoring target address regions')
    parser.add_argument('--numa_node', metavar='<node id>', type=int,
            help='if target is \'paddr\', limit it to the numa node')

def set_implicit_target_monitoring_argparser(parser):
    set_common_monitoring_argparser(parser)
    parser.add_argument('target', type=str, metavar='<target>',
            help='the target (command, pid, or special keywords) to monitor')
    set_common_argparser(parser)

def set_implicit_target_record_argparser(parser):
    set_implicit_target_monitoring_argparser(parser)
    parser.add_argument('-l', '--rbuf', metavar='<len>', type=int,
            help='length of record result buffer')
    parser.add_argument('-o', '--out', metavar='<file path>', type=str,
            default='damon.data', help='output file path')

def set_implicit_target_schemes_argparser(parser):
    set_implicit_target_monitoring_argparser(parser)
    parser.add_argument('-c', '--schemes', metavar='<file or schemes in text>',
            type=str, default='damon.schemes',
            help='data access monitoring-based operation schemes')

def set_explicit_target_argparser(parser):
    set_common_monitoring_argparser(parser)
    parser.add_argument('--ops', choices=['vaddr', 'paddr', 'fvaddr'],
            default='paddr',
            help='monitoring operations set')
    parser.add_argument('--target_pid', type=int, help='target pid')
    parser.add_argument('-c', '--schemes', metavar='<file or schemes in text>',
	    default=json.dumps([_damon.Damos().to_kvpairs()], indent=4),
	    help='data access monitoring-based operation schemes')
    parser.add_argument('--kdamonds', metavar='<string or file>',
            help='key-value pairs format kdamonds config')
    set_common_argparser(parser)



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

"""
Apply given operation schemes.
"""

import argparse
import os
import signal
import subprocess
import time

import _damon
import _damon_args
import _damo_paddr_layout

def cleanup_exit(exit_code):
    kdamonds_names_to_turn_off = []
    if kdamonds_names != None:
        for kdamond_name in kdamonds_names:
            if _damon.is_kdamond_running(kdamond_name):
                kdamonds_names_to_turn_off.append(kdamond_name)
    err = _damon.turn_damon_off(kdamonds_names_to_turn_off)
    if err:
        print('failed to turn damon off (%s)' % err)
    err = _damon.apply_kdamonds(orig_kdamonds)
    if err:
        print('failed restoring previous kdamonds setup (%s)' % err)
    exit(exit_code)

def sighandler(signum, frame):
    print('\nsignal %s received' % signum)
    cleanup_exit(signum)

def set_argparser(parser):
    _damon_args.set_implicit_target_schemes_argparser(parser)

def main(args=None):
    global orig_kdamonds
    global kdamonds_names

    if not args:
        parser = argparse.ArgumentParser()
        set_argparser(parser)
        args = parser.parse_args()

    _damon.ensure_root_and_initialized(args)

    orig_kdamonds = _damon.current_kdamonds()
    kdamonds_names = None

    signal.signal(signal.SIGINT, sighandler)
    signal.signal(signal.SIGTERM, sighandler)

    err, kdamonds = _damon_args.turn_implicit_args_damon_on(args)
    if err:
        print('could not turn DAMON on (%s)' % err)
        cleanup_exit(-3)

    kdamonds_names = [k.name for k in kdamonds]

    print('Press Ctrl+C to stop')
    if args.self_started_target == True:
        os.waitpid(kdamonds[0].contexts[0].targets[0].pid, 0)
    # damon will turn it off by itself if the target tasks are terminated.
    _damon.wait_current_kdamonds_turned_off()

    cleanup_exit(0)

if __name__ == '__main__':
    main()
